!******************************************************************************!
!                                                                              !
!    class_tree.f90                                                            !
!    is part of RECOLA2 (REcursive Computation of One Loop Amplitudes)2        !
!                                                                              !
!    Copyright (C) 2016,2017 Ansgar Denner, Jean-Nicolas Lang and              !
!                            Sandro Uccirati                                   !
!                                                                              !
!    RECOLA2 is licenced under the GNU GPL version 3,                          !
!    see COPYING for details.                                                  !
!                                                                              !
!******************************************************************************!
!                                                           .---.              !
!                                                          /  .  \             !
!                                                         |\_/|   |            !
!                                                         |   |  /|            !
!          .----------------------------------------------------' |            !
!         /  .-.   class_tree.f90                                 |            !
!        |  /   \  This code was generated with REPT1L.           |            !
!        | |\_.  | Change this file and you get what you deserve. |            !
!        |\|  | /| Author: J.N. Lang                              |            !
!        | `---' | UFO date: 07/2017                              |            !
!        |       | UFO version: 1.0                               |            !
!        |       | Generation date: 2018-01-16 00:29             /             !
!        |       |----------------------------------------------'              !
!        \       |                                                             !
!         \     /                                                              !
!          `---'                                                               !
!******************************************************************************!
module class_tree
  use class_vertices
  use constants_mdl
  use tcSUUa, only: treeSUUa
  use tcFFSa, only: treeFFSa
  use tcVVV, only: treeVVV
  use tcSV, only: treeSV
  use tcSSVV, only: treeSSVV
  use tcSSSS, only: treeSSSS
  use tcUUVVa, only: treeUUVVa
  use tcVV, only: treeVV
  use tcFFa, only: treeFFa
  use tcSSV, only: treeSSV
  use tcSSUUa, only: treeSSUUa
  use tcSVV, only: treeSVV
  use tcVVVV, only: treeVVVV
  use tcUUVa, only: treeUUVa
  use tcSSS, only: treeSSS
  use tcFFVa, only: treeFFVa
  use tcSS, only: treeSS
  implicit none

  interface compute_tree_mdl
    module procedure computeTree2, computeTree3, computeTree4, computeTree5,   &
    computeTree6, computeTree7, computeTree8
  end interface 

  contains

  subroutine computeTree2(ty, co, w1, P1, PL1, den, m, wout, last)
    integer, intent(in) :: ty
    logical :: VB_last
    logical, intent(in) :: last
    complex(kind=dp), dimension(0:), intent(in) :: co
    complex(kind=dp), intent(in), dimension(0:3) :: P1, PL1
    complex(kind=dp), intent(in), dimension(0:3) :: w1
    complex(kind=dp), intent(in) :: m
    complex(kind=dp), intent(in) :: den
    complex(kind=dp), dimension(0:3), intent(out) :: wout
    character(len=3) :: ty_str
    complex(kind=dp), dimension(0:3) :: pl, p

    wout(:) = 0.0d0
    VB_last = .false.

    select case(ty)
    case(SV0, VS0)
      call treeSV(ty, co, w1, P1, PL1, den, m, wout, last, VB_last)
    case(VV0, VV1)
      call treeVV(ty, co, w1, P1, PL1, den, m, wout, last, VB_last)
    case(FaF0, FaF1, FaF2, FaF3, FFa0, FFa1, FFa2, FFa3)
      call treeFFa(ty, co, w1, P1, PL1, den, m, wout, last, VB_last)
    case(SS0, SS1)
      call treeSS(ty, co, w1, P1, PL1, den, m, wout, last, VB_last)
    case default
      write(ty_str,'(i3)') ty
      call error_mdl('Unhandled case: ' // ty_str, where='compute_tree_mdl')
    end select 
    if (last .eqv. .false.) then
      wout(:) = cima * wout(:)/den
    else
      if (VB_last) wout(0) = - wout(0)
    end if 
  end subroutine computeTree2

  subroutine computeTree3(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last)
    integer, intent(in) :: ty
    logical :: VB_last
    logical, intent(in) :: last
    complex(kind=dp), dimension(0:), intent(in) :: co
    complex(kind=dp), intent(in), dimension(0:3) :: P1, PL1
    complex(kind=dp), intent(in), dimension(0:3) :: w1
    complex(kind=dp), intent(in), dimension(0:3) :: P2, PL2
    complex(kind=dp), intent(in), dimension(0:3) :: w2
    complex(kind=dp), intent(in) :: m
    complex(kind=dp), intent(in) :: den
    complex(kind=dp), dimension(0:3), intent(out) :: wout
    character(len=3) :: ty_str
    complex(kind=dp), dimension(0:3) :: pl, p

    wout(:) = 0.0d0
    VB_last = .false.

    select case(ty)
    case(UUaS0, SUUa0, USUa0, UaSU0, UaUS0, SUaU0)
      call treeSUUa(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,      &
      VB_last)
    case(FFaS0, FFaS1, FFaS2, FSFa0, FSFa1, FSFa2, FaFS0, FaFS1, FaFS2, SFFa0, &
    SFFa1, SFFa2, SFaF0, SFaF1, SFaF2, FaSF0, FaSF1, FaSF2)
      call treeFFSa(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,      &
      VB_last)
    case(VVV0, VVV1)
      call treeVVV(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,       &
      VB_last)
    case(SVS0, VSS0, VSS1, SSV0, SSV1)
      call treeSSV(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,       &
      VB_last)
    case(VVS0, SVV0, VSV0)
      call treeSVV(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,       &
      VB_last)
    case(UUaV0, UUaV1, UaVU0, VUaU0, VUaU1, VUUa0, VUUa1, UVUa0, UaUV0, UaUV1)
      call treeUUVa(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,      &
      VB_last)
    case(SSS0)
      call treeSSS(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,       &
      VB_last)
    case(FFaV0, FFaV1, VFaF0, VFaF1, FaVF0, FaVF1, FVFa0, FVFa1, FaFV0, FaFV1, &
    VFFa0, VFFa1)
      call treeFFVa(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,      &
      VB_last)
    case default
      write(ty_str,'(i3)') ty
      call error_mdl('Unhandled case: ' // ty_str, where='compute_tree_mdl')
    end select 
    if (last .eqv. .false.) then
      wout(:) = cima * wout(:)/den
    else
      if (VB_last) wout(0) = - wout(0)
    end if 
  end subroutine computeTree3

  subroutine computeTree4(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, den, m&
  , wout, last)
    integer, intent(in) :: ty
    logical :: VB_last
    logical, intent(in) :: last
    complex(kind=dp), dimension(0:), intent(in) :: co
    complex(kind=dp), intent(in), dimension(0:3) :: P1, PL1
    complex(kind=dp), intent(in), dimension(0:3) :: w1
    complex(kind=dp), intent(in), dimension(0:3) :: P2, PL2
    complex(kind=dp), intent(in), dimension(0:3) :: w2
    complex(kind=dp), intent(in), dimension(0:3) :: P3, PL3
    complex(kind=dp), intent(in), dimension(0:3) :: w3
    complex(kind=dp), intent(in) :: m
    complex(kind=dp), intent(in) :: den
    complex(kind=dp), dimension(0:3), intent(out) :: wout
    character(len=3) :: ty_str
    complex(kind=dp), dimension(0:3) :: pl, p

    wout(:) = 0.0d0
    VB_last = .false.

    select case(ty)
    case(SVVS0, VVSS0, VSVS0, SVSV0, SSVV0, VSSV0)
      call treeSSVV(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, den, m, wout&
      , last, VB_last)
    case(SSSS0)
      call treeSSSS(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, den, m, wout&
      , last, VB_last)
    case(UaVUV0, VUaUV0, VVUUa0, UUaVV0, VUUaV0, UaVVU0, UVUaV0, VVUaU0, VUVUa0&
    , UaUVV0, UVVUa0, VUaVU0)
      call treeUUVVa(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, den, m,    &
      wout, last, VB_last)
    case(UaSSU0, USSUa0, SUUaS0, SSUUa0, SUaSU0, UaUSS0, UaSUS0, SUaUS0, SUSUa0&
    , SSUaU0, USUaS0, UUaSS0)
      call treeSSUUa(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, den, m,    &
      wout, last, VB_last)
    case(VVVV0, VVVV1)
      call treeVVVV(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, den, m, wout&
      , last, VB_last)
    case default
      write(ty_str,'(i3)') ty
      call error_mdl('Unhandled case: ' // ty_str, where='compute_tree_mdl')
    end select 
    if (last .eqv. .false.) then
      wout(:) = cima * wout(:)/den
    else
      if (VB_last) wout(0) = - wout(0)
    end if 
  end subroutine computeTree4

  subroutine computeTree5(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, w4, P4&
  , PL4, den, m, wout, last)
    integer, intent(in) :: ty
    logical :: VB_last
    logical, intent(in) :: last
    complex(kind=dp), dimension(0:), intent(in) :: co
    complex(kind=dp), intent(in), dimension(0:3) :: P1, PL1
    complex(kind=dp), intent(in), dimension(0:3) :: w1
    complex(kind=dp), intent(in), dimension(0:3) :: P2, PL2
    complex(kind=dp), intent(in), dimension(0:3) :: w2
    complex(kind=dp), intent(in), dimension(0:3) :: P3, PL3
    complex(kind=dp), intent(in), dimension(0:3) :: w3
    complex(kind=dp), intent(in), dimension(0:3) :: P4, PL4
    complex(kind=dp), intent(in), dimension(0:3) :: w4
    complex(kind=dp), intent(in) :: m
    complex(kind=dp), intent(in) :: den
    complex(kind=dp), dimension(0:3), intent(out) :: wout
    character(len=3) :: ty_str
    complex(kind=dp), dimension(0:3) :: pl, p

    wout(:) = 0.0d0
    VB_last = .false.

    select case(ty)
    case default
      write(ty_str,'(i3)') ty
      call error_mdl('Unhandled case: ' // ty_str, where='compute_tree_mdl')
    end select 
    if (last .eqv. .false.) then
      wout(:) = cima * wout(:)/den
    else
      if (VB_last) wout(0) = - wout(0)
    end if 
  end subroutine computeTree5

  subroutine computeTree6(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, w4, P4&
  , PL4, w5, P5, PL5, den, m, wout, last)
    integer, intent(in) :: ty
    logical :: VB_last
    logical, intent(in) :: last
    complex(kind=dp), dimension(0:), intent(in) :: co
    complex(kind=dp), intent(in), dimension(0:3) :: P1, PL1
    complex(kind=dp), intent(in), dimension(0:3) :: w1
    complex(kind=dp), intent(in), dimension(0:3) :: P2, PL2
    complex(kind=dp), intent(in), dimension(0:3) :: w2
    complex(kind=dp), intent(in), dimension(0:3) :: P3, PL3
    complex(kind=dp), intent(in), dimension(0:3) :: w3
    complex(kind=dp), intent(in), dimension(0:3) :: P4, PL4
    complex(kind=dp), intent(in), dimension(0:3) :: w4
    complex(kind=dp), intent(in), dimension(0:3) :: P5, PL5
    complex(kind=dp), intent(in), dimension(0:3) :: w5
    complex(kind=dp), intent(in) :: m
    complex(kind=dp), intent(in) :: den
    complex(kind=dp), dimension(0:3), intent(out) :: wout
    character(len=3) :: ty_str
    complex(kind=dp), dimension(0:3) :: pl, p

    wout(:) = 0.0d0
    VB_last = .false.

    select case(ty)
    case default
      write(ty_str,'(i3)') ty
      call error_mdl('Unhandled case: ' // ty_str, where='compute_tree_mdl')
    end select 
    if (last .eqv. .false.) then
      wout(:) = cima * wout(:)/den
    else
      if (VB_last) wout(0) = - wout(0)
    end if 
  end subroutine computeTree6

  subroutine computeTree7(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, w4, P4&
  , PL4, w5, P5, PL5, w6, P6, PL6, den, m, wout, last)
    integer, intent(in) :: ty
    logical :: VB_last
    logical, intent(in) :: last
    complex(kind=dp), dimension(0:), intent(in) :: co
    complex(kind=dp), intent(in), dimension(0:3) :: P1, PL1
    complex(kind=dp), intent(in), dimension(0:3) :: w1
    complex(kind=dp), intent(in), dimension(0:3) :: P2, PL2
    complex(kind=dp), intent(in), dimension(0:3) :: w2
    complex(kind=dp), intent(in), dimension(0:3) :: P3, PL3
    complex(kind=dp), intent(in), dimension(0:3) :: w3
    complex(kind=dp), intent(in), dimension(0:3) :: P4, PL4
    complex(kind=dp), intent(in), dimension(0:3) :: w4
    complex(kind=dp), intent(in), dimension(0:3) :: P5, PL5
    complex(kind=dp), intent(in), dimension(0:3) :: w5
    complex(kind=dp), intent(in), dimension(0:3) :: P6, PL6
    complex(kind=dp), intent(in), dimension(0:3) :: w6
    complex(kind=dp), intent(in) :: m
    complex(kind=dp), intent(in) :: den
    complex(kind=dp), dimension(0:3), intent(out) :: wout
    character(len=3) :: ty_str
    complex(kind=dp), dimension(0:3) :: pl, p

    wout(:) = 0.0d0
    VB_last = .false.

    select case(ty)
    case default
      write(ty_str,'(i3)') ty
      call error_mdl('Unhandled case: ' // ty_str, where='compute_tree_mdl')
    end select 
    if (last .eqv. .false.) then
      wout(:) = cima * wout(:)/den
    else
      if (VB_last) wout(0) = - wout(0)
    end if 
  end subroutine computeTree7

  subroutine computeTree8(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, w4, P4&
  , PL4, w5, P5, PL5, w6, P6, PL6, w7, P7, PL7, den, m, wout, last)
    integer, intent(in) :: ty
    logical :: VB_last
    logical, intent(in) :: last
    complex(kind=dp), dimension(0:), intent(in) :: co
    complex(kind=dp), intent(in), dimension(0:3) :: P1, PL1
    complex(kind=dp), intent(in), dimension(0:3) :: w1
    complex(kind=dp), intent(in), dimension(0:3) :: P2, PL2
    complex(kind=dp), intent(in), dimension(0:3) :: w2
    complex(kind=dp), intent(in), dimension(0:3) :: P3, PL3
    complex(kind=dp), intent(in), dimension(0:3) :: w3
    complex(kind=dp), intent(in), dimension(0:3) :: P4, PL4
    complex(kind=dp), intent(in), dimension(0:3) :: w4
    complex(kind=dp), intent(in), dimension(0:3) :: P5, PL5
    complex(kind=dp), intent(in), dimension(0:3) :: w5
    complex(kind=dp), intent(in), dimension(0:3) :: P6, PL6
    complex(kind=dp), intent(in), dimension(0:3) :: w6
    complex(kind=dp), intent(in), dimension(0:3) :: P7, PL7
    complex(kind=dp), intent(in), dimension(0:3) :: w7
    complex(kind=dp), intent(in) :: m
    complex(kind=dp), intent(in) :: den
    complex(kind=dp), dimension(0:3), intent(out) :: wout
    character(len=3) :: ty_str
    complex(kind=dp), dimension(0:3) :: pl, p

    wout(:) = 0.0d0
    VB_last = .false.

    select case(ty)
    case default
      write(ty_str,'(i3)') ty
      call error_mdl('Unhandled case: ' // ty_str, where='compute_tree_mdl')
    end select 
    if (last .eqv. .false.) then
      wout(:) = cima * wout(:)/den
    else
      if (VB_last) wout(0) = - wout(0)
    end if 
  end subroutine computeTree8

    subroutine helicity_conservation_mdl(fheT, typeT, fhfilter, j, w, wOut, ml &
    , last_tl)
      integer, dimension(0:), intent(inout) :: fheT
      logical, dimension(:), intent(inout) :: fhfilter
      integer, intent(in), dimension(:) :: w
      integer, intent(in) :: wOut, typeT, j
      logical, intent(in) :: ml, last_tl
      integer, parameter :: pos_hel = 1
      integer, parameter :: neg_hel = -1
      integer, parameter :: unpol = 0
      integer, parameter :: zero_current = 3
      integer, parameter :: not_set = 111

      select case(typeT)
      case(FFaS0)
        if (fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. unpol .and. fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. unpol) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else
          fheT(wOut) = unpol
        end if 
      case(FFaS1)
        if (fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else
          fheT(wOut) = unpol
        end if 
      case(FFaS2)
        if (fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. unpol) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. unpol .and. fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else
          fheT(wOut) = unpol
        end if 
      case(FFaV0)
        if (fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else
          fheT(wOut) = unpol
        end if 
      case(FFaV1)
        if (fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. unpol) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. unpol .and. fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else
          fheT(wOut) = unpol
        end if 
      case(FaF0)
        fheT(wOut) = unpol
      case(FaF1)
        if (ml .and. fheT(w(1)) .eq. unpol) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(ml .and. fheT(w(1)) .eq. pos_hel) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(fheT(w(1)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else
          fheT(wOut) = unpol
        end if 
      case(FaF2)
        fheT(wOut) = unpol
      case(FaF3)
        fheT(wOut) = unpol
      case(VFaF0)
        if (ml .and. fheT(w(2)) .eq. pos_hel) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(ml .and. fheT(w(2)) .eq. neg_hel) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(VFaF1)
        if (ml .and. fheT(w(2)) .eq. pos_hel) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(ml .and. fheT(w(2)) .eq. unpol) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(FSFa0)
        if (ml .and. fheT(w(1)) .eq. pos_hel) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(ml .and. fheT(w(1)) .eq. neg_hel) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(FSFa1)
        if (ml .and. fheT(w(1)) .eq. unpol) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(ml .and. fheT(w(1)) .eq. pos_hel) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(fheT(w(1)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else
          fheT(wOut) = unpol
        end if 
      case(FSFa2)
        if (ml .and. fheT(w(1)) .eq. unpol) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(fheT(w(1)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(ml .and. fheT(w(1)) .eq. neg_hel) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(FaVF0)
        if (ml .and. fheT(w(1)) .eq. pos_hel) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(ml .and. fheT(w(1)) .eq. neg_hel) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(FaVF1)
        if (ml .and. fheT(w(1)) .eq. unpol) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(ml .and. fheT(w(1)) .eq. pos_hel) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(fheT(w(1)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else
          fheT(wOut) = unpol
        end if 
      case(FaFS0)
        if (fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else
          fheT(wOut) = unpol
        end if 
      case(FaFS1)
        if (fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. unpol .and. fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. unpol) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else
          fheT(wOut) = unpol
        end if 
      case(FaFS2)
        if (fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. unpol) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. unpol .and. fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else
          fheT(wOut) = unpol
        end if 
      case(FVFa0)
        if (ml .and. fheT(w(1)) .eq. pos_hel) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(ml .and. fheT(w(1)) .eq. neg_hel) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(FVFa1)
        if (ml .and. fheT(w(1)) .eq. unpol) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(fheT(w(1)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(ml .and. fheT(w(1)) .eq. neg_hel) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(FFa0)
        fheT(wOut) = unpol
      case(FFa1)
        if (ml .and. fheT(w(1)) .eq. unpol) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(fheT(w(1)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(ml .and. fheT(w(1)) .eq. neg_hel) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(FFa2)
        fheT(wOut) = unpol
      case(FFa3)
        fheT(wOut) = unpol
      case(FaFV0)
        if (fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else
          fheT(wOut) = unpol
        end if 
      case(FaFV1)
        if (fheT(w(1)) .eq. unpol .and. fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. unpol) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else
          fheT(wOut) = unpol
        end if 
      case(SFFa0)
        if (ml .and. fheT(w(2)) .eq. pos_hel) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(ml .and. fheT(w(2)) .eq. unpol) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(SFFa1)
        if (ml .and. fheT(w(2)) .eq. pos_hel) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(ml .and. fheT(w(2)) .eq. neg_hel) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(SFFa2)
        if (fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(ml .and. fheT(w(2)) .eq. neg_hel) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(ml .and. fheT(w(2)) .eq. unpol) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(SFaF0)
        if (ml .and. fheT(w(2)) .eq. pos_hel) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(fheT(w(2)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(ml .and. fheT(w(2)) .eq. unpol) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(SFaF1)
        if (ml .and. fheT(w(2)) .eq. pos_hel) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(ml .and. fheT(w(2)) .eq. neg_hel) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(SFaF2)
        if (fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(ml .and. fheT(w(2)) .eq. neg_hel) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(ml .and. fheT(w(2)) .eq. unpol) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(FaSF0)
        if (ml .and. fheT(w(1)) .eq. pos_hel) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(ml .and. fheT(w(1)) .eq. neg_hel) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(FaSF1)
        if (ml .and. fheT(w(1)) .eq. unpol) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(ml .and. fheT(w(1)) .eq. pos_hel) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(fheT(w(1)) .eq. neg_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else
          fheT(wOut) = unpol
        end if 
      case(FaSF2)
        if (ml .and. fheT(w(1)) .eq. unpol) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(fheT(w(1)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(ml .and. fheT(w(1)) .eq. neg_hel) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(VFFa0)
        if (ml .and. fheT(w(2)) .eq. pos_hel) then
          fheT(wOut) = pos_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(ml .and. fheT(w(2)) .eq. neg_hel) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case(VFFa1)
        if (fheT(w(2)) .eq. pos_hel) then
          fhfilter(j) = .false.
          if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
        else if(ml .and. fheT(w(2)) .eq. neg_hel) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else if(ml .and. fheT(w(2)) .eq. unpol) then
          fheT(wOut) = neg_hel
          if (last_tl) then
            fheT(wOut) = - fheT(wOut)
          end if 
        else
          fheT(wOut) = unpol
        end if 
      case default
        fheT(wOut) = unpol
      end select 
    end subroutine helicity_conservation_mdl
end module class_tree
